cd c:\Python\basic\basic1
# basic1.py
#-------------------------------------------------------------------
# Pythonの基本操作について学ぶ（その1）
# 
#-------------------------------------------------------------------
# 2019/08/14:	新規作成
# 2019/08/19:	コメントなどの訂正と変数名（list）の変更
#-------------------------------------------------------------------
# 予め、Pythonで使用するライブラリを呼び出しておく
import pandas as pd
import numpy as np

#-------------------------------------------------------------------
# １．Pythonのデータ形式
# 本ブログで使用するデータ型のみ説明する
# 複素数型、タプル型、バイト型、集合型等は説明しない
# 変数は、Pythonで定義されたオブジェクトであり、性質や操作できる関数が決まっている
# 尚、ここでは各種括弧について、以下の様に記載する
# () 括弧
# [] 角括弧
# {} 波括弧
#
# １）整数型の例
i = 100
i  # 確認する
j = 200
j  # 確認する
k = i + j
k  # 確認する
k?

# ２）浮動小数点数型の例
x = 100.0
x  # 確認する
y = 1.23
y  # 確認する
z = x * y
z  # 確認する
z?
x = 1.2345e6
x  # 確認する
x?

# ３）文字列型の例
s1 = "qwert"
s1  # 確認する
s2 = "Kuwait"
s2  # 確認する
s3 = s1 + " is not " + s2
s3  # 確認する
#　添え字による部分参照（添え字は0から始まる事に注意）
s2[0]  # K：s2の1文字目
s2[1]  # u
s2[2]  # w
s2[3]  # a
s2[4]  # i
s2[5]  # t
s2[6]  # エラー
#　添え字による後ろからの参照
s3[-1]  # t
s3[-2]  # i
s3[-3]  # a
s3[-4]  # w
s3[-5]  # u
s3[-6]  # K
# 添え字による部分参照（0からの開始位置：終了の次の位置）
s2[0:6]  # Kuwait
len(s3)  # 文字サイズ（19）を確認
s3[13:19]  # Kuwait
# 文字列の置き換え
s1 = "tiger"
s1  # 確認する
s1[0]  # 確認する
s1[0] = "T"  # エラーになる（文字列は、部分的な置き換えができない）
s1 = "Tiger"  # 変数全体の置き換えであれば可能
s1  # 確認する

# ４）リスト型の例（複数の要素を1つの変数内に角括弧を使用し順番を指定して格納）
listx = [1,2,3,"JAPAN",1.2345e2,100.0]  # 6個の異なる型の要素を格納
listx  # 確認する
listx[3]  # 確認する
listx[3] = "Japan"  # 添え字による部分的な置き換えが可能
listx  # 確認する
listx.append(7)  # リスト型の変数オブジェクトに、要素の追加操作を行うappend関数を適用（ピリオドで繋ぐ）
listx  # 7が最後に追加される
listx.clear()  # 要素を削除するclear関数を適用（関数には必ず括弧が付く）
listx  # 確認する
list1 = [1,2,3,4]
list2 = [10,20,30,40]
list3 = [100,200,300,400]
listx = [list1,list2,list3]
listx  # 2次元（3行4列）の配列になっている。尚、多次元配列も可能である
listx[0]  # 1行目を表示
listx[1][3]  # 2行4列の要素を確認
listx[-1]  # 最後の行を確認する
listx = [1,2,3,4,5,6,7,8,9,10]
listx[3]  # 確認する
listx[3:8]  # 複数の要素を部分的に取り出すことも可能
listx?
len(listx)  # len関数で要素数を確認

# ５）辞書型の例（キーと値の組合せにより、順不同で波括弧内に指定する）
dic1 = {1:"01",2:"02",3:"03",4:"04",5:"05",6:"06",7:"07",8:"08",9:"09",0:"00"}
dic1  # 確認する（キーで取り出すため、順番に意味はない）
dic1[5]  # 要素の指定は角括弧で指定
dic1[10]  # 10というキーは存在しないのでエラーになる
dic1?
dic2 = {"apple":[120,125,130,125],"orange":[90,85,95],"banana":[50,40,45,50,55]}
dic2["orange"]
dic2["banana"][4]
dic2  # 確認する

#-------------------------------------------------------------------
# ２．NumPyの配列
# Pandasは、NumPyを元にして開発されている
# NumPyは、多次元配列やベクトル演算を可能にするオブジェクト(ndarray)を提供してくれる
# 但し、ndarrayの多次元配列要素は、全て同じ型のデータでなければならない
# 先ず、Pythonのリスト型の変数を定義する
list1 = [1,2,3,4]
# import numpy as npにより、numpyで定義されている関数を"np."の接頭語を付すことで参照
array1 = np.array(list1)  # リスト型の変数からnumpyのarray関数により、ndarray（NumPyの配列）を作成
array1  # 確認する
array1?
list2 = [10,20,30,40]
list3 = [100,200,300,400]
listx = [list1,list2,list3]
listx  # 確認する
array34 = np.array(listx)  # 3行4列の行列データが作られる
array34  # 確認する
listx * 2  # listの配列が2個作られる
array34 * 2  # array34の要素毎に2が乗ぜられる
# つまり、ndarrayは、行列データであり、行列式の計算ができるオブジェクトである
# 本ブログでは、ndarrayの詳しい説明はしないが、PandasがNumPyを元に作成されており、
# Pandasのデータとndarrayは親和性が高く、Pandasで扱えないデータ処理があった場合、
# ndarrayのデータにして、計算するという事がある。例えば、Pandasに備わっていない
# 機械学習の計算等は、ndarrayのデータに変換してから行う等である
# 本ブログでは、高度な数学や統計を使用する説明は行わない（当面は）ので、本件は無視しても良い。
# 本ブログでは、Pandasで扱える処理を中心に説明する
#
# ここで覚えてもらいたいことは、array34の要素毎に2が乗ぜられるという仕組みについてである
# これは、ブロードキャストと呼ばれる手法で、テレビやラジオの放送と同様に、
# 1カ所で行った処理が、多くの視聴者に同様に届けられるという事を意味している
# array34 * 2は、2を乗ずるという処理が全ての要素に適用されるという事である
#
# また、Pandasは、NumPyを元に開発されていることからNumPyの関数が流用できる

#-------------------------------------------------------------------
# ３．Pandasのデータ
# Series(1次元配列)、DataFrame（2次元配列）の2種類のオブジェクトがある
# １）Seriesは、1つの変数が複数の値を持っている1次元配列である
#     NumPyと同様、全ての要素は、同じ型のデータでなければならない
#     int64：64ビット整数、float64:64ビット浮動小数点数、object：文字列（リスト型等を含む）
#     の3種類を本ブログでは主に使用する。特別な値として日付・時刻のデータ等もある。
#     本ブログでは、積極的に活用する訳ではないが、DataFrame（2次元配列）の集計結果が
#     Seriesで表される為、本ブログでもよく登場することになる
#     例えば、2次元配列の縦横計、平均値等を計算した結果は、1次元になるからである
ser1 = pd.Series([1,2,3,4,5,6,7,8,9,0])  # 全て整数の要素が定義されている
ser1  # int64の値が10個格納されており、0から9までのindexが自動的に付されている
ser2 = pd.Series([1,2,3,4,5,6,7,8,9,0.0])  # 最後の要素が、0.0の浮動小数点で定義されている
ser2  # float64の値が10個格納されており、0から9までのindexが自動的に付されている
ser3 = pd.Series([1,2,3,4,5,6,7,8,9,0],index=[0,1,2,3,4,5,6,7,8,9])  # indexを明示的に設定
ser3  # int64の値が10個格納されており、1から10までのindexが設定されている
ser3?
# オブジェクトのデータタイプは、Seriesであることがわかる
ser3[3]  # 添え字を使って部分参照ができる
ser3[3] = 30
ser3  # 部分的に置き換わっている女とが確認できる
ser3[[3,4,5]]  # 複数の要素を添え字を使って参照できる
# 添え字の範囲を指定して参照（スライス）
ser3[0:5]  # 複数の値を添え字の範囲を指定して参照[start:end]endの1つ前までを参照
ser3[0:5:2]  # 参照する値の間隔をして入りて参照[start:stop:step]stepで間隔を指定
ser3[ser3 > 5]  # 条件を指定した参照、5より大きな要素を抽出
ser3.max()  # 最大値を抽出する関数max()を適用
ser3.sum()  # 合計値を抽出
ser3.mean()  # 平均値を抽出
ser3.describe()  # 記述統計量を算出したSeriesを抽出
ser3.describe()["50%"]  # 中央値のデータを抽出

# ２）DataFrameは、複数の変数の複数の値を扱う2次元配列である
#     各変数の全ての要素は、同じ型のデータでなければならない
#     本ブログのデータは、基本的にDataFrameを中心に扱う
#     また、DataFrameは、列方向と行方向のいずれでも計算が可能であるが、
#     本ブログでは、列方向に変数を配置し、行方向に観察結果を配置する
#     例えば、算数、国語、社会のテストの結果がそれぞれ6回行われたデータであれば、
#     列方向に教科の種類が並び、行方向にテスト日又はテスト順が並ぶことになる
#     列は、縦に並ぶのでcolumns（カラム：柱）と呼び、行は、indexと呼ぶ
#     pandasでは、複数には複数形（columns等）で表すが、indexesとはしない。
df1 = pd.DataFrame({"ser1":ser1,"ser2":ser2,"ser3":ser3})
# 本ブログでは、分析対象データ（csvファイル）を読み込んむ先がDataFrameになる。
# ここでは、説明用に
df1  # ser1-3の変数に対して、0-9の観察結果のデータが配置された行列データが確認できる
df1["ser1"]   # 変数を指定して確認する（int64）
df1["ser2"]   # 変数を指定して確認する（float64）
df1["ser1"][9] = 10   # 添え字を指定して部分的に置き換え => 警告文が表示される
df1  # しかし、ser1の9の値は、10に置き換えられている。
# 警告が気にならなければ、この方法でも良いが、処理時間が掛かるらしいので注意が必要
df1.loc[9,"ser3"] = 10   # ser3の9の値は置き換えられ、警告は表示されない
df1  # ser3も同様に10に置き換えられている
# loc関数で明示的に指定した方が処理も速い。列名行名で指定する場合は、loc関数を使うが、
# 列番号行番号で指定しなければならない時は、iloc関数を使用する
df1.loc[7:9,"ser2"] = 100.0  # 複数の値を選択（スライス）して置き換え
# 複数の参照（スライス）は、start:stop:stepで指定するが、
# locで行ラベル・列ラベルを使う場合はstopまで含まれるが、
# ilocで行番号・列番号を使う場合はstopの一つ前までになるので注意。
df1["ser4"] = df1["ser3"] * 2.0  # 新たな変数を既存の変数から生成（整数 * 浮動小数点数 = 浮動小数点数）
df1  # ser4が追加されている
df1["ser5"] = 0.0
df1  # ser5が追加されて、全要素に0.0が代入されている
df1["ser6"][0] = 1.0  # ser6は、存在しない為エラーになる
df1.loc[0,"ser6"] = 1.0  # ser6の0に1.0を設定し、1～9に対しては、何もしない
df1  # ser6の1～9には、NaN（欠損値：Not a Number）が設定されている。欠損値は欠測値ともいう
df1.index  # df1の行情報が表される
df1.columns  # df1のれる情報が表される
df1.values  # df1の配列値が票kされる
df1.ser1  # df1のser1の配列値が表示される
# index、columns、values、ser1は、ｄｆ１の属性
df1.max()  # df1の変数毎の最大値が表示される
# max()は、括弧付で示されているとおり関数である。属性も関数も予め指定されていなければ指定できない
# DataFrameオブジェクトは、予め複数の属性や関数が設定されている
df1.ser1.dtypes  # ser1のデータ型を確認することができる。
df1.max().max()  # DataFrameの関数maxの結果のSeriesに対して、Seriesの関数maxを実行した結果が表示される
# ビッグデータの場合、行が多量になり、データの確認が容易ではない。
# そこで、最初の5行、最後の3行の様にして確認することが多い（行数は任意）
df1.head(5)  # 最初の5行を表示
df1.tail(3)  # 最後の3行を表示
# 列名が多数存在する場合、列名を縦に並べて確認したい場合は、転置属性（T）を活用する
df1.head(3).T  # 最初の3行を転置して表示する

# ３）欠損値（NaN）を含むDataFrameの取り扱い上の注意
df1["ser1"].dtype  # 整数型であることを確認
df1.loc[5,"ser1"] = np.nan  # 欠損値（np.nan：NumPyで定義されたNaN）を設定
df1["ser1"]  # 浮動小数点数型に変わったことを確認
# 整数型には、欠損値を混在させることができないので、浮動小数点数として扱われる
# csvファイルから読み込んだ場合であっても、欠損値がある整数変数は、浮動小数点数として扱われることに注意する
# オブジェクト型や浮動小数点数型は、欠損値があっても問題はない。
df2 = pd.DataFrame(dic2)  # 辞書型変数の要素数が異なるので定義できない。エラーになる。
dic2  # 確認する
# 欠損値（np.nan：NumPyで定義されたNaN）を指定して、要素数を揃える
dic3 = {"apple":[np.nan,120,125,130,125],"orange":[np.nan,90,85,np.nan,95],"banana":[50,40,45,50,55]}
df3 = pd.DataFrame(dic3,index=["a","b","c","d","e"])  # indexを指定
df3  # 欠損値がある変数は、浮動小数点数になっている
df3.describe()  # 欠損値を無視して記述統計量が計算されていることを確認する
# pandasでは、欠損値は無視して計算されるが、結果として変数毎の件数が合わないことが起こったり
# そもそも欠損値があること自体が問題となることがある
# ビッグデータでは、欠損値の有無を確認するのも容易ではない
df3.isnull().sum()  # DataFrameの要素毎に欠損値か否かを判定(isnull)し、真（True）の数を合計（sum）する
# describeでデータ件数は把握できるが、この方法であれば、変数毎の欠損値の数が容易に分かる
df3["melon"] = np.nan
df3  # 確認する
df3.dropna(how="all",axis=1)  # how="all"は、全てがNaNである場合に削除することを意味し、axis=1は、列方向にチェックすることを意味する
df3.dropna(how="any",axis=1)  # how="any"は、１つでもNaNである場合に削除することを意味し、bananaのみが残ることになる
df4 = df3.dropna(how="all",axis=1)  # 全てが欠損値のmelonを削除したdf4を作成
df4.dropna(how="any",axis=0)  # 1つでも欠損値がある行を削除し、bceだけが残った
df5 = df4.fillna(0)  # 欠損値を0に置き換える
df5  # NaNが0,0に置き換わっていることが分かる
# DataFrameに関数で操作しても置き換えない限りDataFrameに変化は無い事に留意する

# ４）DataFrameで多次元配列を扱う方法
# columnsやindexを多次元にすることは可能であるが、本ブログでは扱わないので説明しない
# しかし、実質的に多次元の分析をすることはあり得る
# 国語_男子、国語_女子、算数_男子、算数_女子の様な変数にすると、実質2次元の変数であるが、変数名は1次元である
# 観察結果も多次元にすることは無いが、データの内容から多次元として扱うことがある
# 変数として、人口、男性人口、女性人口、高齢化率、男性高齢化率、女性高齢化率、世帯数があるようなデータの場合で、
# 市町村別の観察結果があった場合、市町村は、都道府県別にまとめることができる
# つまり、indexとしては、1次元の観察番号となっていても、変数として、都道府県番号や市町村番号があれば、
# 実質的には、都道府県ごとの市町村の分析が可能であり、都道府県単位に集約した分析も可能になる
# 本ブログでは、columnsやindexの多次元化は行わないが、columns（変数名）を工夫することで対応するという事である
# columnsやindexの多次元化は、見にくいので扱わないという事である

